#pragma once

#include <cmath>
#include <jtest/jtest.hpp>
#include <jtest/Unit.hpp>
#include <J3ML/LinearAlgebra/Quaternion.hpp>
#include <J3ML/Algorithm/RNG.hpp>
#include <J3ML/Math.hpp>

jtest::Unit QuaternionUnit {"Quaternion"};
namespace QuaternionTests {

    // This is here to check the accuracy of the Slerp inside the Quaternion class.
    // Although you don't jtest::check anything :shrug: - Redacted.
    Quaternion PreciseSlerp(const Quaternion &a, const Quaternion& b, float t)
    {
        double angle = a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;

        double sign = 1.0;

        if (angle > 0) {
            angle = -angle;
            sign = -1.0;
        }
        double A;
        double B;

        if (angle <= 1.0) // perform spherical linear interpolation
        {
            angle = std::acos(angle); // After this, angle is in the range pi / 2 -> 0 as the original angle variable ranged from 0 -> 1.

            double angleT = t * angle;

            double s[3] = { std::sin(angle), std::sin(angle - angleT), std::sin(angleT)};
            double c = 1.0 / s[0];

            A = s[1] * c;
            B = s[2] * c;
        } else { // If angle is close to taking the denominator to zero, resort to linear interpolation (and normalization).
            A = 1.0 - t;
            B = t;
        }

        Quaternion C;
        C.x = (float)(a.x*A*sign + b.x*B);
        C.y = (float)(a.y*A*sign + b.y*B);
        C.z = (float)(a.z*A*sign + b.z*B);
        C.w = (float)(a.w*A*sign + b.w*B);
        return C.Normalized();
    }

    inline void Define()
    {
        using namespace jtest;
        using namespace J3ML::LinearAlgebra;
        using J3ML::Algorithm::RNG;

        QuaternionUnit += Test("SlerpPrecision", [] {
            RNG rng;
            float maxError = 0;
            float maxLerpError = 0;
            float magnitudeError = 0;
            for (int i = 0; i < 10000; ++i)
            {
                Quaternion q = Quaternion::RandomRotation(rng);
                Quaternion q2 = Quaternion::RandomRotation(rng);
                float t = rng.Float01Incl();

                Quaternion correct = PreciseSlerp(q, q2, t);
                Quaternion fast = q.Slerp(q2, t);

                magnitudeError = std::max(magnitudeError, std::abs(1.f - fast.LengthSquared()));
                Quaternion lerp = q.Lerp(q2, t);
            }
        });

        QuaternionUnit += Test("Constants", [] {
            Quaternion id {0.f, 0.f, 0.f, 1.f};

            jtest::check(id.Equals(Quaternion::Identity));
        });

        QuaternionUnit += Test("From_AxisAngle", [] {
            Quaternion expected_result(0.0579133, 0.0782044, 0.1765667, 0.9794664);
            AxisAngle a({0.2872573, 0.3879036, 0.8757934}, 0.4059981);
            Quaternion from_axis(a);

            jtest::check(Math::EqualAbs(expected_result.x, from_axis.x, 1e-6f));
            jtest::check(Math::EqualAbs(expected_result.y, from_axis.y, 1e-6f));
            jtest::check(Math::EqualAbs(expected_result.z, from_axis.z, 1e-6f));
            jtest::check(Math::EqualAbs(expected_result.w, from_axis.w, 1e-6f));
        });

        QuaternionUnit += Test("Ctor_FromMatrix3x3", [] {
            // TODO: Test multiple rotation values.
            // https://www.andre-gaschler.com/rotationconverter/
            Matrix3x3 matrix_rep = {
                0.9902160,  0.0000000,  0.1395431,
                0.1273699,  0.4084874, -0.9038334,
                -0.0570016,  0.9127640,  0.4044908};
            Quaternion expected = {0.5425029, 0.0586955, 0.0380374, 0.8371371};

            Quaternion from_mat = Quaternion(matrix_rep);

            jtest::check(expected.Equals(from_mat));
        });

        QuaternionUnit += Test("Ctor_FromMatrix4x4", [] {
            Matrix4x4 matrix_rep = {
                0.9799671,  0.1991593, -0.0000000, 0,
                -0.1977032,  0.9728020, -0.1207050, 0,
                -0.0240395,  0.1182869,  0.9926884, 0,
                0, 0, 0, 0};

            Quaternion expected {0.0601595, 0.0060513, -0.0998991, 0.9931588};

            Quaternion q (matrix_rep);

            jtest::check(q.Equals(expected));
        });
        QuaternionUnit += Test("MulOpQuat", [] {
            //Quaternion a =
        });
        QuaternionUnit += Test("DivOpQuat", [] { throw("Not Implemented"); });
        QuaternionUnit += Test("Lerp", [] {
            Quaternion a = Quaternion::RotateX(Math::PiOverTwo);
            Quaternion b = Quaternion::RotateX(-Math::PiOverTwo);

            Quaternion expected {0,0,0,0};

            Quaternion result = a.Lerp(b, 0.5f);
        });
        QuaternionUnit += Test("RotateFromTo", [] { throw("Not Implemented"); });
        QuaternionUnit += Test("Transform", [] { throw("Not Implemented"); });
    }

    inline void Run()
    {
        QuaternionUnit.RunAll();
    }
}



